/*

  Copyright 2010, 2011 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package pado.api;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.util.StringTokenizer;

import org.opt4j.common.archive.ArchiveModule;
import org.opt4j.common.logger.LoggerModule;
import org.opt4j.core.Archive;
import org.opt4j.core.Individual;
import org.opt4j.core.Objectives;
import org.opt4j.core.Population;
import org.opt4j.core.optimizer.Completer;
import org.opt4j.core.optimizer.Optimizer;
import org.opt4j.operator.mutate.BasicMutateModule;
import org.opt4j.optimizer.ea.EvolutionaryAlgorithmModule;
import org.opt4j.start.Opt4JTask;
import org.opt4j.viewer.Viewer;
import org.opt4j.viewer.ViewerModule;

import cf.api.ApplicationEvent;
import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;

import com.google.inject.Module; 

public abstract class PADOApp extends CyberApp
{
  int id;

  /** CONSTURCTOR **/
  public PADOApp(CyberNode node, int nodeIndex)
  {
    super(node);
    this.nodeIndex = nodeIndex;
	this.id = -1;
  }

  public PADOApp(CyberNode node, int nodeIndex, int id)
  {
    super(node);
    this.nodeIndex = nodeIndex;
	this.id = id;
  }

  public int id() { return id; }
  public Archive archive() { return archive; }

  /* Optimizers and miscs */
  EvolutionaryAlgorithmModule ea; /* Genetic Algorithm */
  BasicMutateModule mu; /* Mutation operator in GA */
  ArchiveModule ar; /* Archive (Pareto front) */
  ViewerModule gui; /* Ptolemy GUI */
  LoggerModule logger; /* opt4J logging module */

  int nodeIndex; // index of island
  Random random = new Random();

  /* For logging */
  FileWriter outLog, outSnapshot; 
  BufferedWriter boutLog, boutSnapshot;

  /* For evolutionary process */
  double startTime, endTime; /* start/end time of evolutionary process */
  CloseEvent lastCloseEvent; /* kernel exit event */
  int totalTakenSnapshot= 0;
  int totalPeriod ;//= GENERATION_LIMIT/PADOParam.SHARING_PERIOD;
  int contdConvGen=0; // the generation that convegence rate is kept 
  int curPeriod=0; // 1 step = sharing period

  /* Convergence value-time table */
  HashMap<String, Double> hmConvergeValue = new HashMap<String, Double>();
  HashMap<String, Double> hmConvergeTime = new HashMap<String, Double>();
  Opt4JTask task; /* Optimizing task */

  /* Opt4j data structures */
  Archive archive;
  Population preArchive = new Population();
  Population population;
  Completer completer;
  Viewer viewer;
  Optimizer optimizer;

  public Collection<Module> modules = new ArrayList<Module>();

  /** MEMBER FUNCS **/
  public abstract void userModuleInit();

  public abstract void refresh(Individual i);
  
  public void initialize()
  {
    println("initialization");
    readParam();
    totalPeriod = PADOParam.GENERATION_LIMIT/PADOParam.SHARING_PERIOD;
    userModuleInit();
    initOpt4J();
  }

  public void handleKnowledge(ApplicationKnowledge knew)
  {
    if(knew instanceof BestIndividualKnowledge && !knew.creator().equals( name()
) )
    {
      //System.out.println(name() + " Foreign inidividual comes..." + " " +
      //time() + " from " + knew.creator());
      BestIndividualKnowledge bik = (BestIndividualKnowledge)knew;
      Individual ind = bik.individual;

      if(ind != null)
        if(population.randomReplace(ind))
        {
          //System.out.println(name() + " replace happen " );
          //        System.out.println(name() + "incomming : " + ind + " "  + "
          //        " + ind.getObjectives());
        }
    }

    if(knew instanceof ConvergenceKnowledge && !knew.creator().equals( name() )
)
    {
      ConvergenceKnowledge ck = (ConvergenceKnowledge)knew;
      hmConvergeValue.put( ck.creator() , ck.convergence);
      hmConvergeTime.put ( ck.creator() , new Double(ck.creationTime()) );
    }
  }

  boolean checkTermination() 
  {
    if(PADOParam.TIME_LIMIT <= time() && PADOParam.TIME_LIMIT > 0)
    {
      println(name() + " MEET TIME LIMIT "+ time());
      log( " TIME LIMIT "+ time());
      return true;
    }

    if(curPeriod >= totalPeriod)
    {
      println(name() + " TOTAL STEP EXPIRED");
      log( " TOTAL STEP EXPIRED");
      return true;
    }

    if(!PADOParam.LS)
    {
      double convRate = population.convergenceRate();
      //println(name() + " Convergence Rate : " + convRate);
      if(PADOParam.SHARING)
      {
        ConvergenceKnowledge ck = new ConvergenceKnowledge(this);
        ck.convergence = new Double(convRate);
        postKnowledge( ck );
        hmConvergeValue.put(ck.creator() , ck.convergence);
        hmConvergeTime.put(ck.creator() , ck.creationTime());

        double convRateTotal = 0;
        int cnt=0;
        for(String s : hmConvergeValue.keySet())
        {
          if( hmConvergeTime.get(s).doubleValue() > (time() -
PADOParam.CONV_KNOWLEDGE_EXPIRE_TIME))
          {
            cnt++;
            convRateTotal += hmConvergeValue.get(s).doubleValue();
          }
        }

        if(cnt!=0)
        {
          println(name() + " Global Convergence Rate: " + (convRateTotal/cnt) +
" among " + cnt);
          if((convRateTotal/cnt) >= PADOParam.CONVERGE_RATE_LIMIT)
          {
            //println(name() + " CONVERGE TO "+CONVERGE_RATE_LIMIT);
            contdConvGen += PADOParam.SHARING_PERIOD;
          }
          else
          {
            contdConvGen=0;
          }
        }
      }
      else
      {
        println(name() + " Convergence Rate : " + (convRate));
        if((convRate) >= PADOParam.CONVERGE_RATE_LIMIT)
        {
          contdConvGen += PADOParam.SHARING_PERIOD;
        }
        else
        {
          contdConvGen=0;
        }
      }
    }
    else
    {
      contdConvGen=0;
    }

    if( contdConvGen >= PADOParam.CONTD_CONV_GEN_LIMIT)
    {
      println(name() + " Converge continues " + contdConvGen + " generations");
      log( " Converge continues " + contdConvGen + " generations");
      log( " SAME ARCHIVES CONTINUES");
      return true;
    }

    return false;
  }

  void printPop(Population pop)
  {
    Iterator iter = pop.iterator();
    int index = 0;
    while(iter.hasNext() )
    {
      Individual i = (Individual)iter.next();
      println(index + " : " + i.getObjectives());
      index++;
    }
  }

  void println(String s)
  {
    if(PADOParam.PRINT == 1)
      System.out.println("\033[3"+(nodeIndex+1)+"m"+s);
  }

  void print(String s)
  {
    if(PADOParam.PRINT == 1)
      System.out.print("\033[3"+(nodeIndex+1)+"m"+s);
  }

  protected void initOpt4J()
  {

    ea = new EvolutionaryAlgorithmModule();

    if(PADOParam.GUI)
      gui = new ViewerModule();
  
    task = new Opt4JTask(false);
    ar = new ArchiveModule();
    mu = new BasicMutateModule();
    logger = new LoggerModule();

    ea.setGenerations(PADOParam.SHARING_PERIOD);
    ea.setAlpha(PADOParam.ALPHA);
    ea.setMu(PADOParam.MU);
    ea.setLambda(PADOParam.LAMDA);
    ea.setCrossoverRate(0.95);

    mu.setMutationRate(PADOParam.MUTATION_RATE);

    ar.setCapacity(PADOParam.ARCHIVE_SIZE);

    //gui.setCloseOnStop(true);

    modules.add(ea);
    modules.add(mu);

    if(PADOParam.GUI)
      modules.add(gui);
    modules.add(ar);
    //modules.add(logger);

    try 
    {
      task.init(modules); // module initialize in task
      task.init_p();
    } catch ( Exception e) 
    {
      e.printStackTrace();
    }

    archive = task.getInstance(Archive.class);
    population = task.getInstance(Population.class);
    completer = task.getInstance(Completer.class);
    optimizer =  task.getInstance(Optimizer.class);
    if(PADOParam.GUI)
    {
      viewer = task.getInstance(Viewer.class);
      viewer.optimizationStarted(optimizer);
    }

    startTime = time();
    logFileOpen();
  }

  void readParam()
  {
    try
    {
      println("Read file");
      FileReader fr = new FileReader("param");
      if(fr == null)
        return;
      BufferedReader br = new BufferedReader( fr );
      String s = br.readLine();
      StringTokenizer st = new StringTokenizer(s);
      PADOParam.NUM_THREADS = Integer.parseInt(st.nextToken()); st = new
StringTokenizer(br.readLine());
      PADOParam.NEIGHBOR_FACILITATOR =  st.nextToken(); st = new
StringTokenizer(br.readLine());
      PADOParam.TIME_LIMIT = Integer.parseInt(st.nextToken()); st = new
StringTokenizer(br.readLine());
      PADOParam.GENERATION_LIMIT = Integer.parseInt(st.nextToken()); st = new
StringTokenizer(br.readLine());
      PADOParam.CONVERGE_RATE_LIMIT = Double.parseDouble(st.nextToken()); st =
new StringTokenizer(br.readLine());
      PADOParam.CONTD_CONV_GEN_LIMIT = Integer.parseInt(st.nextToken()); st =
new StringTokenizer(br.readLine());
      PADOParam.SHARING_PERIOD = Integer.parseInt(st.nextToken()); st = new
StringTokenizer(br.readLine());
      PADOParam.ALPHA = Integer.parseInt(st.nextToken() ); st = new
StringTokenizer(br.readLine());
      PADOParam.MU = Integer.parseInt(st.nextToken() ); st = new
StringTokenizer(br.readLine());
      PADOParam.LAMDA = Integer.parseInt(st.nextToken() ); st = new
StringTokenizer(br.readLine());
      PADOParam.MUTATION_RATE = Double.parseDouble(st.nextToken()); st = new
StringTokenizer(br.readLine());
      PADOParam.SHARING_PRBOBABILITY = Double.parseDouble(st.nextToken()); st =
new StringTokenizer(br.readLine());
      PADOParam.CONV_KNOWLEDGE_EXPIRE_TIME = Integer.parseInt(st.nextToken());
st = new StringTokenizer(br.readLine());
      PADOParam.SINGLE_STATE_ITERATION = Integer.parseInt(st.nextToken()); st =
new StringTokenizer(br.readLine());
      PADOParam.M = Integer.parseInt(st.nextToken()); st = new
StringTokenizer(br.readLine());
      PADOParam.ARCHIVE_SIZE = Integer.parseInt(st.nextToken()); st = new
StringTokenizer(br.readLine());
      if(Integer.parseInt(st.nextToken()) == 1)
        PADOParam.GUI = true;
      else
        PADOParam.GUI = false;
      st = new StringTokenizer(br.readLine());
      PADOParam.SNAPSHOT_PERIOD = Integer.parseInt(st.nextToken()); br.close();
      println("NUM_THREADS: " + PADOParam.NUM_THREADS);
      println("NEIGHBOR_FACILITATOR: " + PADOParam.NEIGHBOR_FACILITATOR);
      println("TIME_LIMIT: " + PADOParam.TIME_LIMIT);
      println("GENERATION_LIMIT: " + PADOParam.GENERATION_LIMIT);
      println("CONVERGE_RATE_LIMIT: " + PADOParam.CONVERGE_RATE_LIMIT);
      println("CONTD_CONV_GEN_LIMIT: " + PADOParam.CONTD_CONV_GEN_LIMIT);
      println("SHARING_PERIOD: " + PADOParam.SHARING_PERIOD);
      println("ALPHA: " + PADOParam.ALPHA);
      println("MU: " + PADOParam.MU);
      println("LAMDA: " + PADOParam.LAMDA);
      println("MUTATION_RATE: " + PADOParam.MUTATION_RATE);
      println("SHARING_PRBOBABILITY: " + PADOParam.SHARING_PRBOBABILITY);
      println("CONV_KNOWLEDGE_EXPIRE_TIME: " +
PADOParam.CONV_KNOWLEDGE_EXPIRE_TIME);
      println("SINGLE_STATE_ITERATION: " + PADOParam.SINGLE_STATE_ITERATION);
      println("M: " + PADOParam.M);
      println("ARCHIVE_SIZE: " + PADOParam.ARCHIVE_SIZE);
      println("SNAPSHOT_PERIOD: " + PADOParam.SNAPSHOT_PERIOD);
    }
    catch(IOException e)
    {
      e.printStackTrace();
    }
  }

  void logFileOpen() 
  {
    try 
    {
      File f = new File("log");
      if(f.isDirectory() )
      {
      }
      else
      {
        f.mkdir();
      }

      int o = 0;
      String str = "";
      while(true) 
      {
        str =
"log/"+o+"_"+name()+"_n"+nodeIndex+"_alpha"+PADOParam.ALPHA+"_mu"+PADOParam.MU;
        f = new File(str);
        if(f.isFile() )
        {
          o++;
        }
        else
        { 
          outLog = new FileWriter(str);
          boutLog= new BufferedWriter(outLog);
          break;
        }
      }
    }
    catch( IOException e ) 
    {
      e.printStackTrace();
    }
  }

  void logFileClose()
  {
    try
    {
      boutLog.write("Time Start\n");
      boutLog.write(""+(endTime-startTime)+"\n");
      boutLog.write("Time End\n");
      boutLog.write("Total Generation Start\n");
      boutLog.write(""+curPeriod*PADOParam.SHARING_PERIOD+"\n");
      boutLog.write("Total Generation  End, total snapshot: "+totalTakenSnapshot
+"\n");
      boutLog.write("ALPAHT MU LAMDA MR : " + PADOParam.ALPHA + " " +
PADOParam.MU + " " + PADOParam.LAMDA + " " + PADOParam.MUTATION_RATE + "\n");
      boutLog.write("SHARING GEN : " + PADOParam.SHARING_PERIOD + "\n");
      boutLog.write("Archive Start "+ archive.size() +" \n");
      boutLog.write(""+archive+"\n");
      boutLog.write("Archive End\n");
      boutLog.close();
    }
    catch(IOException e)
    {
      e.printStackTrace();
    }
  }

  void takeSnapshot()
  {
    try
    {
      File f = new File("snapshot");
      if(f.isDirectory() )
      {
      }
      else
      {
        f.mkdir();
      }

      int o = 0;
      String str = "";
      while(true) 
      {
        str =
"snapshot/"+o+"_"+name()+"_n"+nodeIndex+"_alpha"+PADOParam.ALPHA+"_mu"+PADOParam.MU+"_snap";//+"_time"+time();
        //str =
        //"snapshot/"+o+"_"+name()+"_n"+nodeIndex+"_alpha"+ALPHA+"_mu"+MU+"_time"+time();
        f = new File(str);
        if(f.isFile() )
        {
          o++;
        }
        else
        { 
          outSnapshot = new FileWriter(str);
          boutSnapshot = new BufferedWriter(outSnapshot) ;
          break;
        }
      }
      boutSnapshot.write("snap time Start\n");
      boutSnapshot.write(""+(time())+"\n");
      boutSnapshot.write("snap time End\n");
      boutSnapshot.write("Total Generation Start\n");
      boutSnapshot.write(""+curPeriod*PADOParam.SHARING_PERIOD+"\n");
      boutSnapshot.write("Total Generation  End\n");
      boutSnapshot.write("ALPAHT MU LAMDA MR : " + PADOParam.ALPHA + " " +
PADOParam.MU + " " + PADOParam.LAMDA + " " + PADOParam.MUTATION_RATE + "\n");
      boutSnapshot.write("SHARING GEN : " + PADOParam.SHARING_PERIOD + "\n");
      boutSnapshot.write("Archive Start "+ archive.size() +" \n");
      boutSnapshot.write(""+archive+"\n");
      boutSnapshot.write("Archive End\n");
      boutSnapshot.close();

    }
    catch(IOException e)
    {
      e.printStackTrace();
    }


  }

  /******************/
  /** INNER ClASS **/
  /******************/
  class CloseEvent extends ApplicationEvent
  {
    CloseEvent(CyberApplication app, double delay)
    {
      super(app, delay);
      this.app = app;
    }
    CyberApplication app;
    public void handle()
    {
      println(" -END- " + app.name() +", node index:  "+  nodeIndex + ", proceeded generation: " + curPeriod*PADOParam.SHARING_PERIOD + " taken time time :"+ (endTime - startTime)+", current time: "+app.time());
      println("final archive size : " + archive.size());

      for( Individual i : archive )
      {
    	  refresh(i);
      }
      
	  PADOAppEndKnowledge end = new PADOAppEndKnowledge(this.app, id);
	  postKnowledge(end); 

	  if (PADOParam.CONT_PADO == false)
	  {
        task.close();
        logFileClose();

        ((PADOApp)app).terminate();
	  }
    }
  }

  public  class kernelEvent extends ApplicationEvent
  {
    public kernelEvent(CyberApplication app, double delay)
    {
      super(app, delay);
      this.app = app;
    }

    CyberApplication app;

    public void handle()
    {
      try 
      {
        preArchive.clear();
        completer.complete(population);
        archive.update(population);
        preArchive.addAll(archive);

        /* optimization process */
        task.execute_p();

        print(name() +  ", gen: " + (curPeriod*PADOParam.SHARING_PERIOD)+", time: " + app.time() + ", # of archive: " + archive.size());//+ " " + archive);
		 /* Multi Ojbective */
        if(PADOParam.M==1)
        {
          println(", archive value: " + archive);
        }
        else
          System.out.println();
        curPeriod ++;

        /* knowledge posting */
        for(Individual individual : archive)
        {
          for( Individual i : archive )
          {
            refresh(i);
          }
          /* Check Sharing Condition */
          if(PADOParam.SHARING)
          {
            if(PADOParam.SHARING_PRBOBABILITY > random.nextDouble() )
            {
              BestIndividualKnowledge bik = new BestIndividualKnowledge(app,
app.time()+ 600);
              bik.individual = individual;
              postKnowledge(bik);
            }
          }
        }

        /* take snapshot */
        if(app.time() > (totalTakenSnapshot+1)*PADOParam.SNAPSHOT_PERIOD)
        {
          println("!!take snapshot!! time: " + app.time());
          takeSnapshot();
          totalTakenSnapshot++;
        }

        if(checkTermination())
        {
          endTime = app.time();
          if(lastCloseEvent==null)
          {
            lastCloseEvent = new CloseEvent(app, 0.00);
            postEvent(lastCloseEvent);
          }
        }
        else
        {
          if(lastCloseEvent!=null)
          {
            lastCloseEvent.cancel();
            lastCloseEvent = null;
          }
          
		  postEvent(new kernelEvent(app, 0.00));
        }
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
  }
}
